// afl_ext: Very fast procedural ocean 
// https://www.shadertoy.com/view/MdXyzX

#define DRAG_MULT 0.38

// Calculates wave value and its derivative, 
// for the wave direction, position in space, wave frequency and time
vec2 wavedx(vec2 position, vec2 direction, float frequency, float timeshift) {
    float x = dot(direction, position) * frequency + timeshift;
    float wave = exp(sin(x) - 1.0);
    float dx = wave * cos(x);
    return vec2(wave, -dx);
}

// Calculates waves by summing octaves of various waves with various parameters
float getwaves(vec2 position, int iterations) {
    float iter = 0.0; // this will help generating well distributed wave directions
    float frequency = 1.0; // frequency of the wave, this will change every iteration
    float timeMultiplier = 1.5; // time multiplier for the wave, this will change every iteration
    float weight = 1.0;// weight in final sum for the wave, this will change every iteration
    float sumOfValues = 0.0; // will store final sum of values
    float sumOfWeights = 0.0; // will store final sum of weights
    for(int i=0; i < iterations; i++) {
        // generate some wave direction that looks kind of random
        vec2 p = vec2(sin(iter), cos(iter));
        // calculate wave data
        vec2 res = wavedx(position, p, frequency, frameTimeCounter * timeMultiplier);

        // shift position around according to wave drag and derivative of the wave
        position += p * res.y * weight * DRAG_MULT;

        // add the results to sums
        sumOfValues += res.x * weight;
        sumOfWeights += weight;

        // modify next octave ;
        weight = mix(weight, 0.0, 0.2);
        frequency *= 1.18;
        timeMultiplier *= 1.07;

        // add some kind of random value to make next wave look random too
        iter += 1232.399963;
    }
  // calculate and return
    return sumOfValues / sumOfWeights;
}

float getnoise(vec2 pos){
    return textureNice(noisetex, pos, vec2(noiseTextureResolution)).r;
}

float noiseWave(vec2 pos){
    float weight = 100.0;
    float speed = 1.5;
    float time = frameTimeCounter / speed;
    pos *= 0.175;

    float noise = 0; 

    noise += getnoise(((pos + 3.0 *(-pos.yx / 1.3)) + vec2(-1.0, 0.72) * time * 0.2)/8.5)*5.5;    
	noise += getnoise(((pos + 3.0 *(-pos.yx / 1.3)) + vec2(-1.0, 0.72) * time * 0.4)/12.5)*10.5;        
    noise += getnoise(((pos + 3.0 *(-pos.yx / 1.3)) + vec2(-1.0, 0.72) * time * 0.6)/20.5)*15.5;
    noise += getnoise(((pos + 3.0 *(-pos.yx / 1.3)) + vec2(-1.0, 0.72) * time * 0.8)/36.4)*20.4;
    
    noise += getnoise(((pos.yx + 3.0 *(pos / 1.3)) + time * 0.4)/12.5)*10.5;
    noise += getnoise(((pos.yx + 3.0 *(pos / 1.3)) + time * 0.6)/12.5)*15.5;
    noise += getnoise(((pos.yx + 3.0 *(pos / 1.3)) + time * 0.8)/24.5)*20.5;
    noise += getnoise(((pos.yx + 3.0 *(pos / 1.3)) + time * 1.0)/48.6)*25.6;
    
    noise /= weight;
   	noise= pow(noise,1.3);

    return noise;
}

float getWaveHeight(vec2 pos, const int quality){
    pos *= 0.75;
    float waveHeight = getwaves(pos,  quality);
    if(isEyeInWater == 1) waveHeight = 1.0 - waveHeight;
    // float waveHeight = noiseWave(pos);
    return mix(1.0, waveHeight, 0.3);
}

// 效果设计-Parallax Mapping视差映射
// https://miusjun13qu.feishu.cn/docx/G17IdiCyhoEd7XxBqJOcb3J1nie
vec2 waveParallaxMapping(vec2 uv, vec3 viewDirTS){
    const float slicesMin = 5.0;
    const float slicesMax = 15.0;
    const float waveHeight = 1.0;
    const int waveNormalParallaxQuality = 8;

    float slicesNum = ceil(lerp(slicesMax, slicesMin, abs(dot(vec3(0, 0, 1), viewDirTS))));
    float dHeight = 1.0 / slicesNum;
    float rayHeight = 1.0 - dHeight;
    vec2 dUV = waveHeight * (viewDirTS.xy / viewDirTS.z) / slicesNum;
    vec2 currUVOffset = -dUV;
    
    float prevHeight = getWaveHeight(uv, waveNormalParallaxQuality);
    float currHeight = getWaveHeight(uv + currUVOffset, waveNormalParallaxQuality);
    
    for(int i = 0; i < slicesNum; ++i){
        if(currHeight > rayHeight){
            break;
        }
        prevHeight = currHeight;
        currUVOffset -= dUV;
        rayHeight -= dHeight;
        currHeight = getWaveHeight(uv + currUVOffset, waveNormalParallaxQuality);
    }

    float currDeltaHeight = currHeight - rayHeight;
    float prevDeltaHeight = rayHeight + dHeight - prevHeight;
    float weight = currDeltaHeight / (currDeltaHeight + prevDeltaHeight);

    vec2 parallaxUV = uv + currUVOffset + weight * dUV;
    return parallaxUV;
}

vec3 getWaveNormal(vec2 uv){
    const float c = 1.0;
    const int waveQuality = 20;
    vec2 du = vec2(0.05, 0.0);
    vec2 dv = vec2(0.0, 0.05);
    float p = getWaveHeight(uv, waveQuality);
    float p_u = getWaveHeight(uv + du, waveQuality);
    float p_v = getWaveHeight(uv + dv, waveQuality);
    float frac_dp_du = c * (p_u - p) / du.x;
    float frac_dp_dv = c * (p_v - p) / dv.y;

    vec3 normal = normalize(vec3(-frac_dp_du, -frac_dp_dv, 1.0));

    return normal;
}